  ///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////
//
// DynamicArray2D.hpp
//
//	DynamicArray2D header file
//
///////////////////////////////////////
  ///////////////////////////////////////
  
/*

Copyright (c) 2005, Tim Burrell
All rights reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef __DynamicArray2D_h
#define __DynamicArray2D_h

#if HAVE_CONFIG_H
#include "config.h"
#endif

#include "GarbageCollector.hpp"
#include "DynamicArray.hpp"
#include "LinkList.hpp"
#include <stdlib.h>
#include <stdio.h>

//////////////////////////////////////////////////////////////////////////////
// Typedefs
///////////////////////////////////////

enum ERowCol {
	ROW,
	COL
};

//////////////////////////////////////////////////////////////////////////////
// Class Definition
///////////////////////////////////////

template <class DataType> class DynamicArray2DNode : public GarbageCollectorManaged {
public:
	int		Row;
	int		Col;
	DataType * 	pData;

	DynamicArray2DNode(int nRow, int nCol, DataType * Data) { Row = nRow; Col = nCol; pData = Data; };
	virtual ~DynamicArray2DNode() {} ;
};

template <class DataType> class DynamicArray2D {
public:
	void		deleteArray();
	int		getCols();
	DataType *	getDataPtr(int row, int col);
	DataType &	getData(int row, int col);
	int		getRows();
	bool 		setData(int row, int col, DataType & data);
	bool		setSize(int rows, int cols);
	
	DynamicArray2D();
	virtual ~DynamicArray2D();
	
private:
	int 		get1DIndexFrom2D(int row, int col);
	bool 		isIndexValid(ERowCol rowcol, int index);

	DynamicArray< DynamicArray2DNode<DataType> * > m_Array;
	LinkList< DynamicArray2DNode<DataType> > m_DataList;
	int		m_Rows;
	int		m_Cols;
	int		m_LastRows;
	int		m_LastCols;
};

////////////////////////////////////////////////////////////////////////////
// Construction
///////////////////////////////////////

template <class DataType> 
DynamicArray2D<DataType>::DynamicArray2D() {
	m_Rows = 0; m_Cols = 0;
	m_LastRows = 0; m_LastCols = 0;
}

template <class DataType> 
DynamicArray2D<DataType>::~DynamicArray2D() {
	deleteArray();
}

////////////////////////////////////////////////////////////////////////////
// Class Body
///////////////////////////////////////

template <class DataType> 
int DynamicArray2D<DataType>::getRows() {
	return m_Rows;
}

template <class DataType> 
int DynamicArray2D<DataType>::getCols() {
	return m_Cols;
}

template <class DataType> 
DataType & DynamicArray2D<DataType>::getData(int row, int col) {
	return *getDataPtr(row, col);
}

template <class DataType> 
DataType * DynamicArray2D<DataType>::getDataPtr(int row, int col) {
	if ( (!isIndexValid(ROW, row)) || (!isIndexValid(COL, col)) )
		return NULL;

	DynamicArray2DNode<DataType> * pNode = m_Array[get1DIndexFrom2D(row, col)];

	//printf("Returning data from [%d, %d] in pos [%d / %d] = %x\n", row, col, get1DIndexFrom2D(row, col), m_Cols * m_Rows, pNode);

	return pNode->pData;
}

template <class DataType> 
bool DynamicArray2D<DataType>::setData(int row, int col, DataType & data) {
	if ( (!isIndexValid(ROW, row)) || (!isIndexValid(COL, col)) )
		return false;

	DynamicArray2DNode<DataType> * pNode = m_Array[get1DIndexFrom2D(row, col)];
	DataType * pData = pNode->pData;
	*pData = data;
	
	return true;
}

template <class DataType> 
void DynamicArray2D<DataType>::deleteArray() {
	// de-allocate array mem
	if (m_DataList.getSize() > 0) {
		m_DataList.moveToHead();
		do {
			//printf("Getting pNode\n");
			DynamicArray2DNode<DataType> * pNode = m_DataList.getData();
			if (!pNode)
				continue;

			//printf("pNode->data = %x", pNode->pData, pNode);
			if (pNode->pData) {
				delete pNode->pData;
				pNode->pData = NULL;
			}
			
			//printf(" pNode = %x\n", pNode);
			delete pNode;
			pNode = NULL;
		} while (m_DataList.moveNext());

		m_DataList.deleteAll(false);
	}

	m_Array.deleteArray();
	
	m_Rows = 0; m_Cols = 0;
	m_LastRows = 0; m_LastCols = 0;
}

template <class DataType> 
bool DynamicArray2D<DataType>::setSize(int rows, int cols) {
	if ( (rows < 0) || (cols < 0) )
		return false;
		
	m_LastRows = m_Rows; m_LastCols = m_Cols;
	m_Rows = rows; m_Cols = cols;

	//printf("DynamicArray2D :: Size Change From [%dx%d] to [%dx%d]\n", m_LastRows, m_LastCols, m_Rows, m_Cols);

	//int pSize = m_LastRows * m_LastCols;
	int cSize = m_Rows * m_Cols;

	m_Array.setSize(cSize);

	// we need to adjust the parts that don't require new elements
	// these will need to be shifted around in order to match their new locations in the array
	if (m_DataList.getSize() > 0) {
		// now let's adjust the array
		m_DataList.moveToHead();
		do {
			DynamicArray2DNode<DataType> * pNode = m_DataList.getData();

			// check to see if the node is still valid
			// if it's not -- KILL IT!
			// NOTE: This is how we shrink the array!
			if ( (pNode->Row >= m_Rows) || (pNode->Col >= m_Cols) ) {
				if ( (pNode) && (pNode->pData) ) {
					delete pNode->pData;
					pNode->pData = NULL;
				}

				if (pNode) {
					delete pNode;
					pNode = NULL;
				}

				m_DataList.deleteCurrent(false);
				m_DataList.movePrev();
				
				continue;
			}

			//printf("Updating Array[%d][%d]\n", pNode->Row, pNode->Col);
			m_Array[get1DIndexFrom2D(pNode->Row, pNode->Col)] = pNode;
		} while (m_DataList.moveNext());
	}
	
	// let's grow the array, and the list
	// first we update the existing rows with new columns
	for (int r = 0; r < m_LastRows; r ++) {
		for (int c = m_LastCols; c < m_Cols; c ++) {
			DataType * pData = new DataType();
			DynamicArray2DNode<DataType> * pNode = new DynamicArray2DNode<DataType>(r, c, pData);
			m_DataList.addTail(pNode);
			//printf("Setting New Data in Array[%d][%d]\n", pNode->Row, pNode->Col);
			m_Array[get1DIndexFrom2D(r, c)] = pNode;
		}
	}
	
	// let's grow the array, and the list
	// now we update the new rows with full columns
	for (int r = m_LastRows; r < m_Rows; r ++) {
		for (int c = 0; c < m_Cols; c ++) {
			DataType * pData = new DataType();
			DynamicArray2DNode<DataType> * pNode = new DynamicArray2DNode<DataType>(r, c, pData);
			m_DataList.addTail(pNode);
			//printf("Setting New Data in Array[%d][%d]\n", pNode->Row, pNode->Col);
			m_Array[get1DIndexFrom2D(r, c)] = pNode;
		}
	}
	
	return true;
}

template <class DataType> 
inline int DynamicArray2D<DataType>::get1DIndexFrom2D(int row, int col) {
	return (row * m_Cols) + col;
}

template <class DataType> 
bool DynamicArray2D<DataType>::isIndexValid(ERowCol rowcol, int index) {
	if (index < 0)
		return false;

	switch (rowcol) {
	case ROW:
		if (index >= m_Rows)
			return false;
		break;
	case COL:
		if (index >= m_Cols)
			return false;
		break;
	}

	return true;
}

#endif // __DynamicArray2D_h

