////////////////////////////////////////////////////////////////////////////
//
// LinkList.hpp
//
//	LinkList header
//
///////////////////////////////////////

/*

Copyright (c) 2001-2005, Tim Burrell
All rights reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef __LinkList_h
#define __LinkList_h

#if HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>

/////////////////////

template <class DataType> 
class LinkListNode {
public:
	LinkListNode	* prev;
	DataType	* data;
	LinkListNode	* next;
};

template <class DataType> 
class LinkList {
public:
	void  		addHead(DataType * NewHead);
	void  		addTail(DataType * NewTail);
	bool  		appendList(LinkList & ListToAppend);
	void  		deleteAll();
	bool  		deleteArray();
	bool  		deleteCurrent();
	bool		deleteNode(int index);
	DataType * 	getArray();
	DataType * 	getData();
	DataType * 	getDataFromIndex(int index, bool ReturnToOriginalIndex = false);
	int  		getPosition();
	int   		getSize();
	bool  		insertNodeAfterIndex(int Index, DataType * pData);
	bool  		isIndexValid(int Index);
	bool  		movePrev();
	bool  		moveNext();
	bool  		moveTo(int index);
	bool  		moveToHead();
	bool  		moveToTail();
	void  		setArrayNeedsUpdating();
	
	DataType * operator [] (long Index) { return getDataFromIndex(Index, false); }	
		
	LinkList();
	virtual ~LinkList();

protected:
	LinkListNode<DataType> * internal_DeleteNode();
	
	int 	m_Pos;
	int 	m_Size;
	
	bool 	m_Sorted;

	LinkListNode<DataType> * m_Head;
	LinkListNode<DataType> * m_Tail;
	LinkListNode<DataType> * m_Current;
	
private:
	bool  setArraySize(int Size);
	bool  setArrayNeedsUpdatingFrom(int Index);
	bool  updateArray();

	DataType * m_Array;
	bool	m_ArrayMode;
	int	m_ArraySize;
	int	m_ArrayNeedsUpdatingFrom;
};

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

template <class DataType> 
LinkList<DataType>::LinkList() {
	m_Head = NULL;
	m_Tail = NULL;
	m_Current = NULL;
	m_Size = 0;
	m_Pos = -1;

	m_Array = NULL;
	m_ArrayMode = false;
	m_ArraySize = 0;
	m_ArrayNeedsUpdatingFrom = -1;

	m_Sorted = false;
}

template <class DataType> 
LinkList<DataType>::~LinkList() {
	deleteAll();
}

template <class DataType> 
inline void LinkList<DataType>::addHead(DataType * NewHead) {
	LinkListNode<DataType> * tmpNode = new LinkListNode<DataType>;
	
	// setup node;
	tmpNode->next = NULL;
	tmpNode->prev = NULL;
	tmpNode->data = NewHead;
	
	if (!m_Head) {
		// no head -- lets set it to the new head :o)
		m_Head = tmpNode;
		
		m_Tail = m_Head;
		m_Current = m_Head;
		m_Pos = 0;
		
		m_Size = 1;
	} else {
		// already have a head, so let's stick in the new node as the new head
		m_Head->prev = tmpNode;
		tmpNode->next = m_Head;
		m_Head = tmpNode;
		
		m_Size++;
	}

	m_Sorted = false;
	setArrayNeedsUpdatingFrom(0);
}

template <class DataType> 
inline void LinkList<DataType>::addTail(DataType * NewTail) {
	LinkListNode<DataType> * tmpNode = new LinkListNode<DataType>;
	if (!tmpNode)
		return;

	// setup node;
	tmpNode->next = NULL;
	tmpNode->prev = NULL;
	tmpNode->data = NewTail;

	if (!m_Head) {
		// no head -- lets set it to the new tail
		m_Head = tmpNode;

		m_Tail = m_Head;
		m_Current = m_Head;
		m_Pos = 0;

		m_Size = 1;
	} else {
		// already have a head, so let's just add the new node onto the tail
		m_Tail->next = tmpNode;
		tmpNode->prev = m_Tail;
		m_Tail = tmpNode;

		m_Size++;
	}

	m_Sorted = false;
	setArrayNeedsUpdatingFrom(m_Size - 1);
}

template <class DataType> 
inline bool LinkList<DataType>::moveNext() {
	if (m_Size == 0) return false;

	if (m_Current->next) {
		m_Current = m_Current->next;
		m_Pos++;

		return true;
	} else
		return false;
}

template <class DataType> 
inline bool LinkList<DataType>::movePrev() {
	if (m_Size == 0) return false;

	if (m_Current->prev) {
		m_Current = m_Current->prev;
		m_Pos--;

		return true;
	} else 
		return false;
}

template <class DataType>
inline DataType * LinkList<DataType>::getArray() {
	m_ArrayMode = true;

	if (!updateArray()) 	{
		deleteArray();

		return NULL;
	}

	return m_Array;
}

template <class DataType>
inline bool LinkList<DataType>::setArrayNeedsUpdatingFrom(int Index) {
	if ( (m_ArrayNeedsUpdatingFrom > Index) || (m_ArrayNeedsUpdatingFrom < 0) )
		m_ArrayNeedsUpdatingFrom = Index;

	return true;
}

template <class DataType>
inline bool LinkList<DataType>::updateArray() {
	if (m_ArrayNeedsUpdatingFrom < 0)
		return true;

	if (!setArraySize(m_Size))
		return false;

	int CurPos = getPosition();

	int lp = m_ArrayNeedsUpdatingFrom;

	if (!moveTo(m_ArrayNeedsUpdatingFrom))
		if (!moveToHead())
			return false;
		else
			lp = 0;

	do {
		m_Array[lp] = *getData();

		lp ++;
	} while (moveNext());

	moveTo(CurPos);

	m_ArrayNeedsUpdatingFrom = -1;

	return true;
}

template <class DataType>
inline bool LinkList<DataType>::setArraySize(int Size) {
	if (Size == m_ArraySize)
		return true;

	else if (Size == 0)
		return deleteArray();

	else if (Size < 0)
		return false;

	void * pRet = realloc(m_Array, sizeof(DataType) * Size);

	m_ArraySize = Size;
	m_Array = (DataType *) pRet;

	return true;	
}

template <class DataType>
bool LinkList<DataType>::deleteArray() {
	if (m_Array) 	{
		free(m_Array);
		m_Array = NULL;
	}

	m_ArrayMode = false;
	m_ArraySize = 0;
	m_ArrayNeedsUpdatingFrom = -1;

	return true;
}

template <class DataType>
inline DataType * LinkList<DataType>::getData() {
	if (m_Size == 0) return NULL;

	if (m_Current)
		return m_Current->data;
	else
		return NULL;
}

template <class DataType> 
int LinkList<DataType>::getSize() {
	return m_Size;
}

template <class DataType> 
inline bool LinkList<DataType>::moveToHead() {
	if (m_Size <= 0)
		return false;

	m_Current = m_Head;
	m_Pos = 0;

	return true;
}

template <class DataType> 
inline bool LinkList<DataType>::moveToTail() {
	if (m_Size <= 0)
		return false;

	m_Current = m_Tail;
	m_Pos = m_Size - 1;

	return true;
}

template <class DataType> 
inline bool LinkList<DataType>::deleteCurrent() {
	LinkListNode<DataType> * Ret = internal_DeleteNode();

	if (Ret)
	{
		m_Sorted = false;
		setArrayNeedsUpdatingFrom(m_Pos);

		return true;
	}

	return false;
}

template <class DataType> 
inline LinkListNode<DataType> * LinkList<DataType>::internal_DeleteNode() {
	if ( m_Current == NULL)
		return NULL;

	if ( (m_Current->data == NULL) && (m_Current->next == NULL) && (m_Current->prev == NULL) )
		return NULL;

	LinkListNode<DataType> * tNodep;
	LinkListNode<DataType> * Ret;
	
	// point to the node to be deleted
	tNodep = m_Current;
	Ret = tNodep;

	// update the pointers to preserve the list
	if (m_Current->prev) 
		m_Current->prev->next = m_Current->next;
	if (m_Current->next)
		m_Current->next->prev = m_Current->prev;

	// if we're deleting the head or the tail, we need to fix them
	if (m_Pos == 0) {
		if (m_Head->next) {
			m_Head = m_Head->next;
			m_Head->prev = NULL;
		} else
			m_Head = m_Tail;

		m_Current = m_Head;
		m_Pos = 0;
	} else if (m_Pos == m_Size - 1) {
		if (m_Tail->prev) {
			m_Tail = m_Tail->prev;
			m_Tail->next = NULL;
		} else
			m_Tail = m_Head;

		m_Current = m_Tail;
		m_Pos = m_Size - 2;
	} else {
		if (m_Current->prev)  		{
			m_Current = m_Current->prev;
			m_Pos--;
		} else if (m_Current->next) {
			m_Current = m_Current->next;
			m_Pos++;
		} else {
			m_Current = m_Head;

			m_Pos = 0;
		}
	}

	// delete the actual node itself to avoid memory leaks
	delete tNodep;

	// adjust the size to match
	m_Size--;

	// if the list is empty, reset the pointers
	if (m_Size == 0) {
		m_Head = NULL;

		m_Tail = NULL;
		m_Current = NULL;
	}

	else if (m_Size == 1) {
		m_Tail = m_Head;
		m_Current = m_Head;
	}

	m_Sorted = false;
	setArrayNeedsUpdatingFrom(m_Pos);

	return Ret;
}

template <class DataType> 
void LinkList<DataType>::deleteAll() {
	moveToHead();
	while (deleteCurrent());
}

/*
 * void MoveTo(int index)
 *
 * -- moves to a node in the list by cyclicly moving towards the desired index
 *
 * -- enhancements: detect whether or not it's faster to jump to the beginning or the
 *	  end and start cycling from there
*/


template <class DataType> 
inline bool LinkList<DataType>::moveTo(int index) {
	if (index == m_Pos)
		return true;

	if ( (index > -1) && (index < m_Size) ) {
		if (index < m_Pos) {
			while (index != m_Pos)  {
				if (!movePrev())
					return false;
			}
		} else if (index > m_Pos) {
			while (index != m_Pos) 	{
				if (!moveNext())
					return false;
			}
		}
	} else
		return false;


	return true;
}

template <class DataType> 
inline int LinkList<DataType>::getPosition() {
	return m_Pos;
}

template <class DataType> 
inline DataType * LinkList<DataType>::getDataFromIndex(int index, bool ReturnToOriginalIndex) {
	if (!this)
		return NULL;

	DataType * ret;
	int CurPos = m_Pos;

	if (!moveTo(index))
		return NULL;

	ret = getData();

	if (ReturnToOriginalIndex)
		moveTo(CurPos);

	return ret;
}

template <class DataType> 
inline bool LinkList<DataType>::deleteNode(int index) {
	int CurPos = getPosition();

	if ( (index < 0) || (index > getSize()) )
		return false;

	if (!moveTo(index))
		return false;

	deleteCurrent();

	m_Sorted = false;
	setArrayNeedsUpdatingFrom(index);

	moveTo(CurPos);

	return true;
}

template <class DataType> 
inline bool LinkList<DataType>::isIndexValid(int Index) {
	if ( (Index > -1) && (Index < getSize()) )
		return true;

	return false;
}

template <class DataType> 
inline bool LinkList<DataType>::insertNodeAfterIndex(int Index, DataType * pData) {
	if (Index < 0) 	{
		addHead(pData);

		return true;
	}

	if (Index >= (getSize() - 1)) 	{
		addTail(pData);

		return true;
	}

	if (!moveTo(Index))
		return false;

	LinkListNode<DataType> * pNode = new LinkListNode<DataType>;

	pNode->data = pData;

	pNode->prev = m_Current;
	pNode->next = m_Current->next;

	m_Current->next->prev = pNode;
	m_Current->next = pNode;

	m_Size ++;
	m_Current = pNode;
	m_Pos = Index + 1;
	
	m_Sorted = false;
	setArrayNeedsUpdatingFrom(Index);
	
	return true;
}

template <class DataType>
void LinkList<DataType>::setArrayNeedsUpdating() {
	setArrayNeedsUpdatingFrom(m_Pos);
}

template <class DataType>
bool LinkList<DataType>::appendList(LinkList &ListToAppend) {
	if (ListToAppend.getSize() == 0)
		return true;

	ListToAppend.moveToHead();
	do {
		addTail(ListToAppend.getData());

	} while (ListToAppend.moveNext());
	
	m_Sorted = false;
	setArrayNeedsUpdatingFrom(0);
	
	return true;
}

#endif // __LinkList_h

