  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//
// ScriptFunction.cpp
//
//	ScriptFunction class
//
///////////////////////////////////////
  ///////////////////////////////////////
  
/*

Copyright (c) 2005, Tim Burrell
All rights reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "ScriptFunction.hpp"
#include "libGizmo/Debug.hpp"

////////////////////////////////////////////////////////////////////////////
// Statics / Type Defs
///////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
// Overloads
///////////////////////////////////////

bool ScriptFunction::executeFunction() {
	Debug::outDebug("ScriptFunction :: ExecuteFunction");
	return true;
}

////////////////////////////////////////////////////////////////////////////
// Construction
///////////////////////////////////////

ScriptFunction::ScriptFunction() {
	m_Function = NULL;
	Arguments = NULL;
}

ScriptFunction::ScriptFunction(void (* Func) (ScriptFunction *), String FuncName, void * Args) {
	m_Function = Func;
	Arguments = Args;
	FunctionName = FuncName;
}

ScriptFunction::~ScriptFunction() {
	resetArgLists();
}

//////////////

ScriptFuncArg::ScriptFuncArg() {
	ArgNumber = 0.0;
	ArgBoolean = false;
}

ScriptFuncArg::~ScriptFuncArg() {
	ArgString.freeString();
}

ScriptFuncArg::ScriptFuncArg(String Value) {
	setString(Value);
}

ScriptFuncArg::ScriptFuncArg(double Value) {
	setNumber(Value);
}

void ScriptFuncArg::setString(String Value) {
	Type = SCRIPT_ARG_STRING;
	ArgString = Value;
	ArgNumber = 0.0;
	ArgBoolean = false;
}

void ScriptFuncArg::setNumber(double Value) {
	Type = SCRIPT_ARG_NUMBER;
	ArgNumber = Value;
	ArgBoolean = false;
}

void ScriptFuncArg::setBoolean(bool Value) {
	Type = SCRIPT_ARG_BOOLEAN;
	ArgNumber = 0.0;
	ArgBoolean = Value;
}

////////////////////////////////////////////////////////////////////////////
// Class Body
///////////////////////////////////////

void ScriptFunction::resetArgLists() {
	resetArgList(&InArgList);
	resetArgList(&OutArgList);
}

void ScriptFunction::resetArgList(LinkList<ScriptFuncArg> * ArgList) {
	if (ArgList->getSize() == 0)
		return;
	ArgList->moveToHead();
	do {
		ScriptFuncArg * pArg = ArgList->getData();
		delete pArg;
	} while (ArgList->moveNext());

	ArgList->deleteAll();
}

void (*ScriptFunction::getFunction(void)) (ScriptFunction *) {
	return m_Function;
}

bool ScriptFunction::checkScriptFunc(ScriptFunction * pFunc) {
	if (!pFunc) {
		Debug::outError("ScriptFunction :: Invalid ScriptFunction!");
		return false;
	}
	if (!pFunc->Arguments) {
		Debug::outError("ScriptFunction :: %s: Invalid Arguments!", pFunc->FunctionName.getString());
		return false;
	}
	return true;
}

ScriptFuncArg * ScriptFunction::addInArg(ScriptFuncArgType Type) {
	ScriptFuncArg * pArg = new ScriptFuncArg;
	pArg->Type = Type;
	InArgList.addTail(pArg);
	
	return pArg;
}

ScriptFuncArg * ScriptFunction::addInArg(ScriptFuncArgType Type, String Value) {
	ScriptFuncArg * pArg = new ScriptFuncArg(Value);
	pArg->Type = Type;
	InArgList.addTail(pArg);
	
	return pArg;
}

ScriptFuncArg * ScriptFunction::addInArg(ScriptFuncArgType Type, double Value) {
	ScriptFuncArg * pArg = new ScriptFuncArg(Value);
	pArg->Type = Type;
	InArgList.addTail(pArg);
	
	return pArg;
}

ScriptFuncArg * ScriptFunction::addOutArg(ScriptFuncArgType Type) {
	ScriptFuncArg * pArg = new ScriptFuncArg;
	pArg->Type = Type;
	OutArgList.addTail(pArg);
	
	return pArg;
}

ScriptFuncArg * ScriptFunction::addOutArg(ScriptFuncArgType Type, String Value) {
	ScriptFuncArg * pArg = new ScriptFuncArg(Value);
	pArg->Type = Type;
	OutArgList.addTail(pArg);
	
	return pArg;
}

ScriptFuncArg * ScriptFunction::addOutArg(ScriptFuncArgType Type, double Value) {
	ScriptFuncArg * pArg = new ScriptFuncArg(Value);
	pArg->Type = Type;
	OutArgList.addTail(pArg);
	
	return pArg;
}

ScriptFuncArg *	ScriptFunction::getInArg(int Index) {
	if (!InArgList.moveTo(Index))
		return NULL;

	return InArgList.getData();
}

ScriptFuncArg *	ScriptFunction::getOutArg(int Index) {
	if (!OutArgList.moveTo(Index))
		return NULL;

	return OutArgList.getData();
}

void ScriptFunction::debugPrintInArgs() {
	// put the arguments into the ScriptFunction
	for (int lp = 0; lp < InArgList.getSize(); lp ++) {
		ScriptFuncArg * pArg = getInArg(lp);
		if (!pArg)
			continue;
		switch (pArg->Type) {
		case SCRIPT_ARG_NUMBER:
			Debug::outDebug("InArgList[%d] == %.2lf", lp, pArg->ArgNumber);
			break;
		case SCRIPT_ARG_STRING:
			Debug::outDebug("InArgList[%d] == %s", lp, pArg->ArgString.getString());
			break;
		case SCRIPT_ARG_BOOLEAN:
			Debug::outDebug("InArgList[%d] == %d", lp, pArg->ArgBoolean);
			break;
		}
	}
}
