  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//
// GizmoPluginX11.cpp
//
//	GizmoPluginX11 class
//
///////////////////////////////////////
  ///////////////////////////////////////
  
/*

Copyright (c) 2005, Tim Burrell
All rights reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "GizmoPluginX11.hpp"
#include "libGizmo/Debug.hpp"
#include "libGizmo/UtilTime.hpp"

#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>
#include <X11/extensions/xf86vmode.h>
#include <X11/keysym.h>

////////////////////////////////////////////////////////////////////////////
// Statics / Type Defs
///////////////////////////////////////

struct gnome_area_info {
    long harea; // current
    long varea; // current
    long harea_count;
    long varea_count;
};

struct gnome_workspace_info {
    long workspace; // current
    long workspace_count;
};

bool GizmoPluginX11::m_ErrorHandlerHit = false;

#define getbit(buf,n) (buf [(n) >> 3] & (1 << ((n) & 7)))

////////////////////////////////////////////////////////////////////////////
// Overrides
///////////////////////////////////////

void GizmoPluginX11::onForegroundWindowChange(String NewWindow) {
	const int IN_ARGS = 3;
	
	if (m_pScriptEngine == NULL)
		return;

	ScriptFuncArg * pInArg[IN_ARGS];
	int lp;
	
	for (lp = 0; lp < IN_ARGS; lp ++) {
		pInArg[lp] = m_ScriptFuncOnX11ForegroundWindowChange.getInArg(lp);
		if (!pInArg[lp]) {
			Debug::outDebug("GizmoPluginX11 :: Failed to Call OnX11ForegroundWindowChange -- Arg %d", lp);
			return;
		}
	}
		
	pInArg[0]->ArgNumber = m_DeviceID;
	pInArg[1]->ArgString = m_DevicePath;
	pInArg[2]->ArgString = NewWindow;
		
	m_pScriptEngine->callScriptFunction(&m_ScriptFuncOnX11ForegroundWindowChange);
}

bool GizmoPluginX11::onPluginOpened() {
	Debug::outDebug("GizmoPluginX11 :: onPluginOpened");
	
	return initPlugin();
}

////////////////////////////////////////////////////////////////////////////
// Callbacks
///////////////////////////////////////

int GizmoPluginX11::x11ErrorHandler(Display * display, XErrorEvent * error) {
	//Debug::outDebug("GizmoPluginX11 :: x11 Error detected!");
	// do nothing we don't care
	m_ErrorHandlerHit = true;
	return 0;
}

int GizmoPluginX11::x11IOErrorHandler(Display * display) {
	//Debug::outDebug("GizmoPluginX11 :: x11 IO Error detected!");
	// do nothing we don't care
	m_ErrorHandlerHit = true;
	return 0;
}

////////////////////////////////////////////////////////////////////////////
// Construction
///////////////////////////////////////

GizmoPluginX11::GizmoPluginX11(ScriptEngineInterface * pScriptEngine, const char * DevicePath) : 
	m_ScriptFuncX11GetCurrentWorkspace(this),
	m_ScriptFuncX11GetNumWorkspaces(this),
	m_ScriptFuncX11SetWorkspace(this),
	m_ScriptFuncX11SetWorkspaceNext(this),
	m_ScriptFuncX11SetWorkspacePrev(this)
{
	Debug::outDebug("GizmoPluginX11 :: X11 Input Plugin Loading...");

	m_pScriptEngine = pScriptEngine;
	m_DevicePath = DevicePath;

	m_Display = NULL;
	m_WatchingDevice = false;
}

GizmoPluginX11::~GizmoPluginX11() {
	closePlugin();
	Debug::outDebug("GizmoPluginX11 :: X11 Input Plugin Unloaded.");
}

////////////////////////////////////////////////////////////////////////////
// Class Body
///////////////////////////////////////

void GizmoPluginX11::closePlugin() {
	Debug::outDebug("GizmoPluginX11 :: X11 Input Plugin Closing...");

	m_WatchingDevice = false;
	
	closeDisplay();
}

bool GizmoPluginX11::initPlugin() {
	Debug::out("        - X11 Input Plugin v%s [%s]", PACKAGE_VERSION, m_DevicePath.getString());

	if (!m_pScriptEngine) {
		Debug::outError("GizmoPluginX11 :: Invalid Script Engine!");
		return false;
	}

	if (!openDisplay(m_DevicePath)) {
		Debug::outError("GizmoPluginX11 :: Failed to Open X11 Display [%s]!", m_DevicePath.getString());
		return false;
	}
	
	// register functions
	Debug::outDebug("GizmoPluginX11 :: Registering Functions with Script Engine...");
	m_pScriptEngine->registerScriptFunction(&m_ScriptFuncX11GetCurrentWorkspace);
	m_pScriptEngine->registerScriptFunction(&m_ScriptFuncX11GetNumWorkspaces);
	m_pScriptEngine->registerScriptFunction(&m_ScriptFuncX11SetWorkspace);
	m_pScriptEngine->registerScriptFunction(&m_ScriptFuncX11SetWorkspaceNext);
	m_pScriptEngine->registerScriptFunction(&m_ScriptFuncX11SetWorkspacePrev);

	Debug::outDebug("GizmoPluginX11 :: X11 Input Plugin Initialized.");
	return true;
}

bool GizmoPluginX11::checkEventsWaiting() {
	//Debug::outDebug("GizmoPluginX11 :: Watching Device Events...");
	
	m_WatchingDevice = true;
	static String LastWin;
	String WinName;
	getForegroundWindowProperties(&WinName);
	if (WinName == LastWin)
		return false;

	// window changed! notify
	onForegroundWindowChange(WinName);
	LastWin = WinName;
	
	return true;
}

bool GizmoPluginX11::openDisplay(const char * DisplayName) {
	if (m_Display)
		closeDisplay();

	if (!DisplayName)
		m_DisplayName = XDisplayName(NULL);
	else
		m_DisplayName = DisplayName;


	if ( (m_Display = XOpenDisplay(m_DisplayName.getString())) == NULL) {
		Debug::outWarn("GizmoPluginX11 :: Unable to Open Display [%s]", m_DisplayName.getString());
		return false;
	}

	// set the error handlers
	// special thanks to ajax from #xorg for this tip
	// you have no idea how annoying this was to track down
	XSetErrorHandler(x11ErrorHandler);
	XSetIOErrorHandler(x11IOErrorHandler);

	m_Screen = DefaultScreen(m_Display);
	m_RootWindow = RootWindow(m_Display, m_Screen);

	Debug::outDebug("GizmoPluginX11 :: Display Opened [%s]", m_DisplayName.getString());
	
	return true;
}

void GizmoPluginX11::closeDisplay() {
 	if (!m_Display)
		return;

	XCloseDisplay(m_Display);
	m_Display = NULL;

	Debug::outDebug("GizmoPluginX11 :: Display Closed");
}

bool GizmoPluginX11::getForegroundWindowProperties(String * WindowName) {
 	if (!m_Display) {
		*WindowName = "(not connected to X11)";
		return false;
	}

	Window 		window;
	int		revert_to_return;

	XGetInputFocus(m_Display, &window, &revert_to_return);
	*WindowName = getWindowName(window);

	//Debug::outDebug("GizmoPluginX11 :: Active Window = [0x%lx] [%s]", window, WindowName->getString());

	return true;
}

/*
 * Routine to display a window id in dec/hex with name if window has one
 * Taken from xwininfo.c from xorg source, thanks xorg guys you rule!
 *
 * I modified it to correctly get the parent window if necessary
 */
 
String GizmoPluginX11::getWindowName(Window window, bool recurse) {
    XTextProperty tp;
    Display * dpy = m_Display;

    if (!window) {
	return "";
    } else {
	if (window == RootWindow(dpy, XDefaultScreen(dpy))) {
	    return "(root window)";
	}
#ifdef NO_I18N
	char * win_name;
	if (!XFetchName(dpy, window, &win_name)) { /* Get window name if any */
	    XFree(win_name);
	    Window root_ret, parent_ret;
	    unsigned int nChildren;
	    Window * children;
	    XQueryTree(dpy, window, &root_ret, &parent_ret, &children, &nChildren);
	    XFree(children);
	    if ((XGetWMName(dpy, parent_ret, &tp)) && (recurse)) {
    		XFree((void*)tp.value);
	    	return getWindowName(parent_ret, false);
            } else {
	    	if (tp.value)
	    		XFree((void*)tp.value);
	        return "(no name)";
	    }
	} else if (win_name) {
	    const String retStr = win_name;
	    XFree(win_name);
	    return retStr;
	}
#else
	if (!XGetWMName(dpy, window, &tp)) { // Get window name if any
    	    if (tp.value)
	    	XFree((void*)tp.value);
	    Window root_ret, parent_ret;
	    unsigned int nChildren;
	    Window * children;
	    XQueryTree(dpy, window, &root_ret, &parent_ret, &children, &nChildren);
	    /* causes seg faults!
	       if ( (children) && (nChildren > 0) )
		    XFree(children);
	    */
	    if ((XGetWMName(dpy, parent_ret, &tp)) && (recurse)) {
  		XFree((void*)tp.value);
	    	return getWindowName(parent_ret, false);
            } else {
	    	if (tp.value)
	    		XFree((void*)tp.value);
	        return "(no name)";
	    }
        } else if (tp.nitems > 0) {
	    String retStr;
            int count = 0, i, ret;
            char **list = NULL;
            ret = XmbTextPropertyToTextList(dpy, &tp, &list, &count);
            if((ret == Success || ret > 0) && list != NULL) {
                for(i=0; i<count; i++)
		    retStr += (const char *) list[i];
    	        XFree((void*)tp.value);
                XFreeStringList(list);
		return retStr;
            } else {
     	    	XFree((void*)tp.value);
	    	if (list)
	                XFreeStringList(list);
		retStr = (const char *) tp.value;
		return retStr;
	    }
	}
#endif
	else
	    return "(no name)";
    }

    return "(unknown)";
}

/*
 * Thanks to the EvRouter project for this one
 */
unsigned int GizmoPluginX11::getKeyboardModifierState(bool * shift, bool * alt, bool * ctrl) {
	int i, j;
	char km [32];
	unsigned char mods = 0;

	*shift = *alt = *ctrl = false;

	if (!m_Display)
		return 0;

	XQueryKeymap(m_Display, km);
	XModifierKeymap * xmk = XGetModifierMapping(m_Display);

	for(j = 0; j < 8; j++) {
		for(i = 0; (i < xmk->max_keypermod) && xmk->modifiermap[(j * xmk->max_keypermod) + i]; i++) {
			if (getbit(km, xmk->modifiermap[(j * xmk->max_keypermod) + i])) {
				mods |= (1 << j);

				switch (j) {
				case 0:
					*shift = true;
					break;
				case 2:
					*ctrl = true;
					break;
				case 3:
					*alt = true;
					break;
				}

			}
		}
	}

	mods &= ~2; /* Ignore the 'Lock' mod */

	XFreeModifiermap(xmk);

	return mods;
}

/*
 * Alt does not currently work with non XTEST
 */
/*
bool GizmoPluginX11::sendKey(int KeyCode, bool Ctrl, bool Alt, bool Shift) {
 	if (!m_Display)
		return false;
	
	Window 		window;
	int		revert_to_return;
	XKeyEvent 	event;
	Window 		root_ret, parent_ret;
	unsigned int 	nChildren;
	Window * 	children;

	XGetInputFocus(m_Display, &window, &revert_to_return);
	XQueryTree(m_Display, window, &root_ret, &parent_ret, &children, &nChildren);
	    
	event.display = m_Display;
	event.window = window;
	event.root = root_ret;
	event.subwindow = None;
	event.time = CurrentTime;
	event.x = 1;
	event.y = 1;
	event.x_root = 1;
	event.y_root = 1;
	event.same_screen = true;
	
	event.keycode = XKeysymToKeycode(m_Display, KeyCode);
	event.state = Shift ? ShiftMask : 0 | Ctrl ? ControlMask : 0;

	// send keypress
	event.type = KeyPress;
	XSendEvent(event.display, event.window, true, KeyPressMask, (XEvent *) &event);

	// send key release
	event.type = KeyRelease;
	XSendEvent(event.display, event.window, true, KeyPressMask, (XEvent *) &event);

	XFlush(m_Display);
	
	return true;
}

bool GizmoPluginX11::sendKeyXTest(int KeyCode, bool Ctrl, bool Alt, bool Shift) {
#ifdef HAVE_XTEST
	int 	kc;
	KeySym	ks;
	int	Ckc = 0;
	int	Akc = 0;
	int	Skc = 0;

	// press mods
	if (Ctrl) {
		ks = XStringToKeysym("Control_L");
		Ckc = XKeysymToKeycode(m_Display, ks);
		XTestFakeKeyEvent(m_Display, Ckc, true, 0);
	}

	if (Alt) {
		ks = XStringToKeysym("Alt_L");
		Akc = XKeysymToKeycode(m_Display, ks);
		XTestFakeKeyEvent(m_Display, Ckc, true, 0);
	}
	
	if (Shift) {
		ks = XStringToKeysym("Shift_L");
		Skc = XKeysymToKeycode(m_Display, ks);
		XTestFakeKeyEvent(m_Display, Skc, true, 0);
	}
	
	kc = XKeysymToKeycode(m_Display, KeyCode);
	XTestFakeKeyEvent(m_Display, kc, true, 0);
	XTestFakeKeyEvent(m_Display, kc, false, 0);

	// release mods
	if (Ckc)
		XTestFakeKeyEvent(m_Display, Ckc, false, 0);
	if (Akc)
		XTestFakeKeyEvent(m_Display, Akc, false, 0);
	if (Skc)
		XTestFakeKeyEvent(m_Display, Skc, false, 0);

	return true;
#else
	return false;
#endif
}

bool GizmoPluginX11::sendMouseButton(int button, bool Pressed) {
 	if (!m_Display)
		return false;

#ifdef HAVE_XTEST
	XTestFakeButtonEvent(m_Display, button, Pressed, CurrentTime);

	Debug::outDebug("GizmoPluginX11 :: Sent Mouse Press [%d] [%d]", button, Pressed);

	return true;
#else
	return false;
#endif
}

// takes zero arguments
// returns 3 keyboard modifiers (shift, alt, ctrl)
void GizmoPluginX11::scriptFunction_X11GetKeyMods(ScriptFunction * pFunc) {
	if (!ScriptFunction::checkScriptFunc(pFunc))
		return;

	GizmoPluginX11 * pThis = (GizmoPluginX11 *) pFunc->Arguments;

	// get keyboard state
	bool shift, alt, ctrl;
	pThis->getKeyboardModifierState(&shift, &alt, &ctrl);

	Debug::outDebug("GizmoPluginX11 :: keyMods = %d%d%d", shift, alt, ctrl);
	
	// return shift
	ScriptFuncArg * pArg = new ScriptFuncArg();
	pArg->Type = SCRIPT_ARG_BOOLEAN;
	pArg->ArgBoolean = shift;
	pFunc->OutArgList.addTail(pArg);

	// return alt
	pArg = new ScriptFuncArg();
	pArg->Type = SCRIPT_ARG_BOOLEAN;
	pArg->ArgBoolean = alt;
	pFunc->OutArgList.addTail(pArg);

	// return ctrl
	pArg = new ScriptFuncArg();
	pArg->Type = SCRIPT_ARG_BOOLEAN;
	pArg->ArgBoolean = ctrl;
	pFunc->OutArgList.addTail(pArg);
}
*/

/*
void GizmoPluginX11::scriptFunction_X11GetActiveWinTitle(ScriptFunction * pFunc) {
	if (!ScriptFunction::checkScriptFunc(pFunc))
		return;

	GizmoPluginX11 * pThis = (GizmoPluginX11 *) pFunc->Arguments;

	// get win props
	String WindowName;
	pThis->getForegroundWindowProperties(&WindowName);

	Debug::outDebug("GizmoPluginX11 :: ActiveWindowName = %s", WindowName.getString());
	
	// return the name
	ScriptFuncArg * pArg = new ScriptFuncArg();
	pArg->Type = SCRIPT_ARG_STRING;
	pArg->ArgString = WindowName;
	pFunc->OutArgList.addTail(pArg);
}
*/

String GizmoPluginX11::getDeviceName() {
	return "GizmoPluginX11";
}

String GizmoPluginX11::getDeviceNameAlt() {
	return m_DevicePath;
}

/*
 * Thanks to the 3ddeskd project for this one (thanks Brad Wasson)!
 * All the stuff the follows is courtesy Brad Wasson
 */
void GizmoPluginX11::get_vdesktop_info(int * column, int * row, int * n_columns, int * n_rows) {
	int c = 0, r = 0, nc = 0, nr = 0;

	Debug::outDebug("vdesktops: EWMH");
	struct gnome_workspace_info wi;
	ewmh_get_desktop_info (&wi);
	c = wi.workspace;
	nc = wi.workspace_count;
	nr = 1;

	Debug::outDebug("c=%d, r=%d, nc=%d, nr=%d", c, r, nc, nr);

	if (column)    *column = c;
	if (row)       *row = r;
	if (n_columns) *n_columns = nc;
	if (n_rows)    *n_rows = nr;
}

/* un-unashamedly ripped from stph.c, thanks Rob :) */
/* unashamedly ripped and adapted from gwmh.c, thanks Tim :) */
/* unashamedly ripped and adapted from 3ddesk, thanks Brad Wasson! */
int GizmoPluginX11::send_client_message_32 (Display *display, Window recipient, Window event_window, Atom message_type, long event_mask, int n_longs, ...) {
    XEvent xevent = { 0, };
    int i = 0;
    va_list var_args = {0};

    if (n_longs > 5)
        return 1;
  
    va_start (var_args, n_longs);

    xevent.type = ClientMessage;
    xevent.xclient.window = event_window;
    xevent.xclient.message_type = message_type;
    xevent.xclient.format = 32;
    while (n_longs--)
        xevent.xclient.data.l[i++] = va_arg (var_args, long);

    va_end (var_args);

    if (!(XSendEvent (display, recipient, False, event_mask, &xevent)))
        return 1;
    if (m_ErrorHandlerHit) {
        m_ErrorHandlerHit = 0;
        return 1;
    }
  
    return 0;
}

#ifndef min
#define min(a,b)  ((a) < (b) ? (a) : (b))
#endif
int GizmoPluginX11::Get_Window_Property(Display *display, Window target_win, Atom atom, long *length, Atom *type, int *size, char **data) {
    Atom actual_type;
    int actual_format;
    unsigned long nitems;
    unsigned long nbytes;
    unsigned long bytes_after;
    unsigned char *prop;
    int status;

    unsigned int max_len = 100;

    status = XGetWindowProperty(display, target_win, atom, 0, (max_len+3)/4, False, AnyPropertyType, &actual_type, &actual_format, &nitems, &bytes_after, &prop);
  
    if (m_ErrorHandlerHit) {
        m_ErrorHandlerHit = 0;
        return 1;
    }

    if (status != Success) /* some other problem... */
        return 1;

    if (actual_format == 32)
        nbytes = sizeof(long);
    else if (actual_format == 16)
        nbytes = sizeof(short);
    else
        nbytes = 1;
    *length = min(nitems * nbytes, max_len);
    *type = actual_type;
    *size = actual_format;
    *data = (char *)prop;

    if (prop == NULL)
        return 1;

    return 0;
}
#undef min

//===========================
// EWMH - desktops (GNOME 2, KDE 3 etc)
//===========================

void GizmoPluginX11::ewmh_set_current_desktop (int workspace) {
    int rc;
    Atom workspace_atom;

    workspace_atom = XInternAtom(m_Display, "_NET_CURRENT_DESKTOP", False);
    rc = send_client_message_32(m_Display,
                                m_RootWindow, m_RootWindow,
                                workspace_atom,
                               (SubstructureNotifyMask |
                                SubstructureRedirectMask),
                                2,
                                workspace,
                                CurrentTime);
    XSync(m_Display, 0);
    return;
}

void GizmoPluginX11::ewmh_get_desktop_info(struct gnome_workspace_info *info)
{
    Atom desktop_atom;
    Atom desktop_count_atom;
    char *val;
    long length;
    Atom type;
    int size;
   
    desktop_atom = XInternAtom(m_Display, "_NET_CURRENT_DESKTOP", False);
    desktop_count_atom = XInternAtom(m_Display, "_NET_NUMBER_OF_DESKTOPS", False);

    if (Get_Window_Property(m_Display, m_RootWindow, desktop_atom, &length, &type, &size, &val)) {
        Debug::outDebug("get property _NET_CURRENT_DESKTOP failed - setting one");
        info->workspace = 0;
    } else {
        info->workspace = (long) * (long *)val;
        XFree(val);
    }

    if (Get_Window_Property(m_Display, m_RootWindow, desktop_count_atom, &length, &type, &size, &val)) {
        Debug::outDebug("get property _NET_NUMBER_OF_DESKTOPS failed - setting one");
        info->workspace_count = 1;
    } else {
        info->workspace_count = (long) * (long *)val;
        XFree(val);
    }

    return;
}

int GizmoPluginX11::getWorkspace() {
	int c, r, nc, nr;
	get_vdesktop_info(&c, &r, &nc, &nr);

	return c;
}

int GizmoPluginX11::getNumWorkspaces() {
	int c, r, nc, nr;
	get_vdesktop_info(&c, &r, &nc, &nr);

	return nc;
}

bool GizmoPluginX11::setWorkspace(int Workspace) {
	// set some stuff
	int c, r, nc, nr;
	get_vdesktop_info(&c, &r, &nc, &nr);
	if ( (Workspace < 0) || (Workspace >= nc) )
		return false;
	ewmh_set_current_desktop(Workspace);
	return true;
}

bool GizmoPluginX11::setWorkspaceNext() {
	// set some stuff
	int c, r, nc, nr;
	get_vdesktop_info(&c, &r, &nc, &nr);
	c ++;	
	if (c >= nc)
		c = 0;
	ewmh_set_current_desktop(c);
	return true;
}

bool GizmoPluginX11::setWorkspacePrev() {
	// set some stuff
	int c, r, nc, nr;
	get_vdesktop_info(&c, &r, &nc, &nr);
	c --;
	if (c < 0)
		c = nc - 1;
	ewmh_set_current_desktop(c);
	return true;
}
